stylevariantpreview widget.wallpaper {
  border-radius: 6px;
  box-shadow: 0 0 9px 1px rgba(0,0,0,.2);
}
stylevariantpreview widget.window {
  border-radius: 6px;
  box-shadow: 0 1px 4px 1px alpha(black, 0.13),
              0 1px 10px 5px alpha(black, 0.09),
              0 3px 16px 8px alpha(black, 0.04),
              0 0 0 1px alpha(black, .05);
}
stylevariantpreview widget.window.dark {
  background: #242424;
}
stylevariantpreview widget.window.light {
  background: #fafafa;
}
stylevariantpreview widget.header {
  border-radius: 7px 7px 0 0;
}
stylevariantpreview widget.header.dark {
  background: #303030;
  box-shadow: inset 0 -1px alpha(black, .36);
}
stylevariantpreview widget.header.light {
  background: #ebebeb;
  box-shadow: inset 0 -1px alpha(black, .07);
}

popover.global-search contents {
  padding: 0;
  margin: 0;
}
popover.global-search contents .navigation-sidebar {
  border-right: 1px solid @borders;
}
popover.global-search listview.results {
  padding: 6px 0;
  background: none;
  color: inherit;
}
popover.global-search listview.results row {
  min-height: 36px;
  padding: 2px 8px;
  border-radius: 6px;
  margin: 0 6px 4px;
}

notificationstack button {
  margin: 0;
  padding: 0;
  background: none;
}

scrolledwindow.undershoot-top > undershoot.top,
preferencespage.undershoot-top > scrolledwindow > undershoot.top {
  box-shadow: inset 0 1px alpha(@shade_color, .75);
  background: linear-gradient(to bottom, alpha(@shade_color, .75), transparent 4px);
}

scrolledwindow.undershoot-bottom > undershoot.bottom,
preferencespage.undershoot-bottom > scrolledwindow > undershoot.bottom {
  box-shadow: inset 0 -1px alpha(@shade_color, .75);
  background: linear-gradient(to top, alpha(@shade_color, .75), transparent 4px);
}

button.pill.small {
  font-size: .83333em;
  border-radius: 99px;
  margin: 0;
  padding: 1px 12px;
}

/* Styling for search bars */
.searchbar {
  background-color: @popover_bg_color;
  background-clip: padding-box;
  color: @popover_fg_color;
  padding: 6px;
  border-style: solid;
  border-color: alpha(black, .14);
  border-left-width: 1px;
  border-right-width: 1px;
  border-bottom-width: 1px;
  border-radius: 0 0 12px 12px;
  margin: 0 10px 20px 10px;
  box-shadow: 0 1px 5px 1px alpha(black, .09),
              0 2px 14px 3px alpha(black, .05);
}
.searchbar button.flat.circular {
  min-width: 24px;
  min-height: 24px;
  margin: 0px;
  padding: 3px;
}

/* Images for checkbuttons in various UI */
.checkimage radio,
.checkimage check {
  background: none;
  outline: none;
  box-shadow: none;
  border: none;
  color: @window_fg_color;
  -gtk-icon-source: none;
  -gtk-icon-size: 18px;
}
.checkimage radio:checked,
.checkimage check:checked {
  -gtk-icon-source: -gtk-icontheme('object-select-symbolic');
}
