/* gbp-flatpak-sources.h
 *
 * Copyright 2016 Endless Mobile, Inc.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

GFile    *gbp_flatpak_sources_fetch_archive (const gchar  *url,
                                             const gchar  *sha,
                                             const gchar  *module_name,
                                             GFile        *destination,
                                             guint         strip_components,
                                             GError      **error);
gboolean  gbp_flatpak_sources_apply_patch   (const gchar  *path,
                                             GFile        *source_dir,
                                             guint         strip_components,
                                             GError      **error);
