.IdeTweaksWindow .checkimage radio,
.IdeTweaksWindow .checkimage check {
  background: none;
  outline: none;
  box-shadow: none;
  border: none;
  color: @window_fg_color;
  -gtk-icon-source: none;
  -gtk-icon-size: 18px;
}
.IdeTweaksWindow .checkimage radio:checked,
.IdeTweaksWindow .checkimage check:checked {
  -gtk-icon-source: -gtk-icontheme('object-select-symbolic');
}
.IdeTweaksWindow button.link {
  padding: 0;
  margin: 0;
}

.IdeTweaksWindow:backdrop IdeTweaksPanelList,
.IdeTweaksWindow:backdrop IdeTweaksPanel {
  color: alpha(currentColor, .5);
}
