window.shortcutui list row.has-header {
  border-top: 1px solid @card_shade_color;
  border-top-left-radius: 0;
  border-top-right-radius: 0;
}
window.shortcutui list row list row.header {
  border-top-left-radius: 0;
  border-top-right-radius: 0;
}
window.shortcutui list > label.heading {
  padding: 12px 12px 12px 14px;
}
